try:
   try:
      # When Python >= 3.8
      from importlib import metadata as importlib_metadata

   except ImportError:
      # When Python < 3.8
      import importlib_metadata

   __version__ = importlib_metadata.version(__package__)

except ImportError:
   __version__ = 'unknown'


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tfmake', 'tfmake.custom']

package_data = \
{'': ['*'], 'tfmake': ['templates/*']}

install_requires = \
['Jinja2>=2.11.1,<3.0.0',
 'PyYAML>=5.3.1,<6.0.0',
 'click>=7.0,<8.0',
 'outdated>=0.2.0,<0.3.0',
 'pathlib>=1.0.1,<2.0.0']

entry_points = \
{'console_scripts': ['tfmake = tfmake.cli:main']}

setup_kwargs = {
    'name': 'tfmake',
    'version': '1.0.52',
    'description': 'Python based Makefile wrapper for terraform projects',
    'long_description': '# tfmake\n\n## Hello ! 👋\n\nThis is a Python based wrapper around an opionated `Makefile` I use for multi-cloud/cross-account `terraform` projects. \n\nYou still need `make`, though. The main advantage is that you don\'t have to copy the `Makefile`. \n\n## DISCLAIMER\nThis module includes a highly opinionated `Makefile` implementation. It\'s working very well for us, but your requirements might be different.\n\nAlso, I\'ve done my best to make this thing work on both MacOS and Linux. If you get an error message (~ and my colleagues haven\'t beat you to it) ... please get in touch and/or create a PR.\n\n## Install\n\n```\n$ pip install tfmake\n```\n\n## Usage\n> Here\'s the help for the wrapper\n```\n$ tfmake --help\nUsage: tfmake [OPTIONS] COMMAND [ARGS]...\n\nOptions:\n  --version  Show the version and exit.\n  --help     Show this message and exit.\n\nCommands:\n  aws    Use AWS provider.\n  azure  Use Azure provider\n  guess  Default command that guesses what provider you\'re using.\n  init   Create configuration for provider.\n```\n\nA warning will be shown when your `tfmake` is out-of-date. If you see one, please update: `pip install --upgrade tfmake` \n\n## Providers\nCurrently, `tfmake` supports two providers: `aws` and `azure`. The default provider is `aws`. Depending on the selected provider, a different, provider specific, `Makefile` is used to wrap `terraform`. Here, the provider is selected by using the right command.\n\nSee \'examples\' for some ... examples.\n\nEach provider leads to a specific `Makefile`. For example: `provider==azure` leads to using `Makefile.azure`.\n\n## Provider Authentication\nThe used `Makefile` will _not_ handle authentication. It just assumes you\'re using an authenticated context.\n\nFor, `aws`, I use [`aws-vault`](https://github.com/99designs/aws-vault). For `azure`, I use the [`azure-cli`](https://docs.microsoft.com/en-us/cli/azure/install-azure-cli?view=azure-cli-latest).\n\n> Here\'s the help for the (bundled) Makefile\n```\n$ tfmake help\nThe AWS Edition\n\nUsage: make <TARGET> (env=<ENVIRONMENT>) (<TERRAFORM ARGUMENTS>)\n\nWhere:\n ENVIRONMENT is one of [\'dev\',\'tst\',\'acc\',\'qas\',\'prd\',\'run\']\n TARGET is one:\n    update                          Update terraform modules and providers\n    select                          Select and initialize terraform workspace (aka \'stage\')\n    show                            Show current terraform workspace\n    plan                            Generate and show an execution plan\n    apply                           Builds or changes infrastructure\n    destroy                         Destroy Terraform-managed infrastructure\n    refresh                         Refresh terraform state\n    import                          Import existing infrastructure into your Terraform state\n\nNote:\n\n - parameter \'env\' is only required when selecting an environment\n - <TERRAFORM ARGUMENTS> can be used to pass arbitrary terraform parameters (without a prefix dash). Example: "make apply input=false no-color auto-approve"\n```\n\n> Use `tfmake azure help` to see the `azure` edition ...\n\n## Workspace Prefix\n\nWhen selecting an environment, by default, the following naming convention is used to store the `terraform` state:\n\n```\n<ACCOUNT_ALIAS>/<TERRAFORM_S3_KEY>\n```\n\nWhere:\n* `ACCOUNT_ALIAS` is the alias of the account you\'re using (either: AWS or Azure)\n* `TERRAFORM_S3_KEY` is the value of the `key` in `terraform.tf`\n\nIf you want something in-between the alias and the key, you can use the \'workspace prefix\' like this:\n> Only needed when using the `select` command!\n\n```\n$ aws-vault exec my-aws-account -- tfmake select env=dev --workspace-key-prefix foo/bar\n```\n\nAfter you `select`ed the environment, the path to your `terraform` state is:\n\n```\n<ACCOUNT_ALIAS>/<WORKSPACE_PREFIX>/<TERRAFORM_S3_KEY>\n```\n\nAssuming the `key` value is `whatever/terraform.tfstate`, this leads to: `my-aws-account/foo/bar/whatever/terraform.tfstate`\n\n\n## Safeguarding Credentials\n> This works for both `AWS` and `Azure`\n\nAlthough, you have to provide the correct credentials yourself, `tfmake` _will_ check if the credentials you provide are expected.\n\n```\n$ cd /path/to/my/tf/project\n$ aws-vault exec ACCOUNT_ONE -- tfmake select env=dev\nInitializing modules...\n...\n>>> Now using Terraform workspace \'dev\' on AWS account \'aws-account-one\' ...\n```\n\nIf you use different credentials the next time you use `tfmake` in this project, you\'ll see a warning. Press `<ENTER>` to abort (or `y` to continue).\n\n```\n$ cd /path/to/my/tf/project\n$ aws-vault exec ACCOUNT_TWO -- tfmake plan\n[WARNING] You previously used \'aws-account-one\' for provider aws. Now you\'re using \'aws-account-two\'. Are you sure? [y/N]:\nAborted!\n```\n\n> The \'magic\' behind all this is a cache stored in `$(PWD)/.tfmake/cache`. So, if you delete this file, this safeguard won\'t work anymore :(\n\n## Auto-switching Credentials on Azure\n\nWhen you\'re using `Azure` _and_ installed [`azctx`](https://github.com/StiviiK/azctx), `tfmake` will use the cached credentials to automatically switch to the correct subscription.\n\nIn your `.tfmake/config`, make sure you enabled `auto_switching`:\n```\n---\nprovider: azure\nauto_switch: True\n\nenvironment:\n  - ARM_ACCESS_KEY = $(az keyvault secret show --name YOUR_SECRET --vault-name YOUR_VAULT --query value -o tsv)\n```\n\n> Support for `AWS` will follow in a future version of `tfmake`\n\n## Final Notes\n\nBy default, before any a `terraform` command is executed, you will be asked to confirm the usage of the current environment.\n\n```\n$ tfmake azure apply\n\nUsing workspace \'prd\' on \'My_fancy_Azure_Production_subscription\'.\n\nPress [ENTER] to continue or [CTRL-C] to stop.\n```\n\nPlease notice that the prompt shows the selected `terraform` workspace and the alias/name of the provider account.\n\n> Use `TFMAKE_AGREE=1` to auto confirm that prompt ...\n\n## Example\n\n> Initialise \'dev\' environment\n```\n$ aws-vault exec foobar -- tfmake select env=dev\n```\n\n> Plan changes\n```\n$ aws-vault exec foobar -- tfmake plan\n```\n\n> Plan changes with plan file\n```\n$ aws-vault exec foobar -- tfmake plan out=foobar.plan\n```\n\n> Apply changes\n```\n$ aws-vault exec foobar -- tfmake apply\n```\n\n> Apply changes using plan file\n```\n$ aws-vault exec foobar -- tfmake apply input=false auto-approve foobar.plan\n```\n\n> Apply changes using the `azure` provider\n```\n$ az login\n# (optionaly set subscription)\n$ az account set --subscription=YOUR_SUBSCRIPTION_ID_HERE\n$ tfmake azure apply\n```\n\n> Apply changes ... automagically\n```\n$ aws-vault exec foobar -- tfmake apply input=false no-color auto-approve\'\n```\n\n**Note**: you can use arbitrary Terraform arguments after the command you\'re executing (~ in this example: `apply`). So, in this example, `input=false no-color auto-approve` will be used to create the following `terraform` arguments: `-input=false -no-color -auto-approve`.\n\nNotice that you have to skip the dash in front of each argument.\n\n**Another Note**: :arrow_up ... this only applies when using `tfmake` ... if, for whatever reason, you choose to use the `Makefile` directly: run `make help`\n\n## Advanced Usage\n\nYou can use a _per_ project configuration file in which you can specify, for example, environment variables and/or arbitrary commands that need to be execute before or after the `make` target.\n\n```\n$ tfmake init aws\n\nConfiguration written to \'/your/current/directory/.tfmake\'\n```\n\n```\n---\n#\n# Welcome to the wonderful world of TFMAKE!\n#\n# This file is the main config file for your tfmake project.\n# It\'s very minimal at this point and uses default values.\n# You can always add more config options for more control.\n#\n# Happy Coding!\n#\nprovider: aws\n\n# You can define service wide environment variables here\n# Notes:\n# - when using commands, dont use single quotes (sorry)\n# - commands should be between $( and )\n#\n# environment:\n#    - variable1 = value\n#    - variable2 = $(command)\n\n# You can define commands here that are executed _before_ the Make target\n# Note: dont use single quotes (sorry)\n#\n# before:\n#    - echo "[INFO] before"\n\n# You can define commands here that are executed _after_ the Make target\n# Note: dont use single quotes (sorry)\n#\n# after:\n#    - echo "[INFO] after"%\n```\n\n#### Section `provider`\nThe `provider` section specifies what (default) provider to use for this project. As a result, you don\'t have to specify it anymore on the command-line.\n\nSo, no more `tfmake azure plan`. Instead, just configure the `provider` property and type `tfmake plan`.\n\n#### Section `environment`\nThe `environment` section can contain a list of environment variables that need to be set before calling the `make` target. \n\nFor example, when using the `azure` provider, you might want to dynamically fetch the value for the `ARM_SAS_TOKEN` or `ARM_ACCESS_KEY` from an Azure Key Vault.\n\n#### Section `before`\n\nThe `before` section can contain a list of commands that need to be executed preparing the call to the `make` target.\n\n#### Section `after`\n\nThe `after` section can contain a list of commands that you can use to cleanup after calling the `make` target.\n\n### Example\n\n> Azure\n\nThe following example configuration creates an environment variable using a specified secret from a given key vault. Then, it will execute all `before` statements. Next it will just execute the `Makefile` target. And finally, it will execute all `after` statements.\n\nPlease note that the `environment` variables will _not_ be exposed system-wide. They will only be \'visible\' within the context of the session. As a result, `terraform` can use it, but when the process is done, the variable will no longer be available.\n\n```\n---\nprovider: azure\n\nenvironment:\n    - ARM_ACCESS_KEY = $(az keyvault secret show --name YOUR_SECRET --vault-name YOUR_VAULT --query value -o tsv)\n\nbefore:\n    - echo $ARM_ACCESS_KEY\n\nafter:\n    - echo "DONE!!"\n```\n\n~ the end\n',
    'author': 'Pascal Prins',
    'author_email': 'pascal.prins@foobar-it.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/paprins/tfmake',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*',
}


setup(**setup_kwargs)



import os
import re
import subprocess
from enum import Enum
import inspect
from functools import wraps
import yaml
import click

from tfmake import __version__
from outdated import check_outdated

class PROVIDER(Enum):
    """
    Supporder providers.
    """
    AWS   = 'aws'
    AZURE = 'azure'

    @classmethod
    def has_value(cls, value):
        return value in cls._value2member_map_ 

    @classmethod
    def to_list(cls):
        return cls._value2member_map_.keys()

    @classmethod
    def to_string(cls):
        return ", ".join(cls._value2member_map_.keys())

class WORKSPACE(Enum):
    """
    Supported workspaces/environments.
    """
    DEV     = 'dev'
    TST     = 'tst'
    ACC     = 'acc'
    QAS     = 'qas'
    PRD     = 'prd'
    RUN     = 'run'
    SANDBOX = 'sandbox'

    @classmethod
    def has_value(cls, value):
        return value in cls._value2member_map_

    @classmethod
    def to_string(cls):
        return ", ".join(cls._value2member_map_.keys())

class DefaultCommandGroup(click.Group):
    '''
    Allow a default command for a group
    '''

    def command(self, *args, **kwargs):
        """
        Extension of default 'click' behavior: run 'default' command if no command is available.
        """
        default_command = kwargs.pop('default_command', False)
        if default_command and not args:
            kwargs['name'] = kwargs.get('name', '<>')
        decorator = super(
            DefaultCommandGroup, self).command(*args, **kwargs)

        if default_command:
            def new_decorator(f):
                cmd = decorator(f)
                self.default_command = cmd.name
                return cmd

            return new_decorator

        return decorator

    def resolve_command(self, ctx, args):
        try:
            # test if the command parses
            return super(
                DefaultCommandGroup, self).resolve_command(ctx, args)
        except click.UsageError:
            # command did not parse, assume it is the default command
            args.insert(0, self.default_command)
            return super(
                DefaultCommandGroup, self).resolve_command(ctx, args)

def check_latest_version(f):
    """
    Wrapper that checks if newer version of 'tfmake' is available.
    """
    @wraps(f)
    def wrapper(self, *args, **kw):
        is_outdated = False
        latest_version = __version__
        try:
            is_outdated, latest_version = check_outdated('tfmake', __version__)

        except ValueError as e:
            click.secho(str(e), bold=True)

        if is_outdated:
            _msg = 'The package tfmake is out of date. Your version is {}, the latest is {}.'.format(__version__, latest_version)
            
            click.secho('\n' + ('* ' * 43), bold=True)
            click.secho(_msg, bold=True)
            click.secho('* ' * 43, bold=True)

        return f(self, *args, **kw)

    return wrapper

def before_and_after(f):
    @wraps(f)
    def wrapper(self, *args, **kw):
        if hasattr(self, 'before') and inspect.ismethod(self.before):
            self.before(*args)
        result = f(self, *args)
        if hasattr(self, 'after') and inspect.ismethod(self.after):
            self.after(*args, **kw)
        return result
    return wrapper

class DefaultCommandHandler(object):
    def __init__(self, provider = None):
        self.config   = None
        self.switched = False

        # filter os.environ for TFMAKE variables
        self.tfmake_env = {key: value for key, value in os.environ.items() if key.startswith('TFMAKE_')}

        base = os.path.join(os.getcwd(), '.tfmake')

        if os.path.isfile(base):
            raise click.ClickException("Legacy configuration found! Run 'tfmake init' first.")

        if not os.path.isdir(base):
            os.mkdir(base)

        _config = os.path.join(base, 'config')
        _cache  = os.path.join(base, 'cache')

        if os.path.isfile(_config):
            with open(_config, 'r') as f:
                try:
                    self.config = yaml.safe_load(f)
                except yaml.parser.ParserError:
                    os.sys.exit('[ERROR] {} could not be parsed.'.format(_config))

        if not provider and self.config:
            self.provider = self.config.get('provider','aws').lower()
        else:
            self.provider = provider or 'aws'

        if self.config:
            _provider = self.config.get('provider','aws').lower()
            if not self.provider == _provider:
                raise click.ClickException("hmmm ... you configured '{}' and provided '{}' ... make up your mind!\n\n\n... please".format(_provider, self.provider))

        # Check if provider is valid
        if not PROVIDER.has_value(self.provider):
            raise click.ClickException("Invalid provider '{}' (expecting: {})".format(self.provider, PROVIDER.to_string()))

        # Check cache
        if os.path.isfile(_cache):
            with open(_cache, 'r') as f:
                try:
                    self.cache = yaml.safe_load(f)
                except yaml.parser.ParserError:
                    os.sys.exit('[ERROR] {} could not be parsed.'.format(_cache))
        else:
            self.cache = dict()

    def __eval(self, return_code):
        """
        Evaluate return return_code from 'os.system'.

        Parameters
        ----------
        return_code: int
            return code from 'os.system'

        Raises
        ------
        ClickException
            when 'return_code' is not equal to '0'
        """
        if return_code != 0:
            raise click.ClickException(return_code)

    def __get_environment(self, target, args=[]):
        """
        Get environment via 'select' target or terraform cli.

        Parameters
        ----------
        target:str
            Makefile target
        args:list(str)
            Arguments for 'Makefile' target.

        Raises
        ------
        ClickException
            When invalid request, or other error.
        """
        if target == "select":
            _args = dict(item.split("=") for item in args) 
            if 'env' in _args:
                environment = _args.get('env')
            else:
                raise click.ClickException("target '{}' expects an argument 'env'.".format(target))
        else:
            try:
                cmd = "terraform workspace show"
                environment = subprocess.check_output(cmd.split(), stderr=subprocess.STDOUT, universal_newlines=True).strip()

            except subprocess.CalledProcessError as e:
                raise click.ClickException("executing '{}': {}".format(cmd, str(e)))

            except OSError as e:
                raise click.ClickException("did you install terraform?")

        if target not in ['help','foo'] and not WORKSPACE.has_value(environment):
            raise click.ClickException("environment '{}' is not supported (expecting: {})".format(environment, WORKSPACE.to_string()))

        return environment

    def __get_account_alias(self):
        """
        Get provider account alias.
        Using cli to avoid adding Python dependencies!
        """
        if PROVIDER(self.provider) == PROVIDER.AWS:
            cmd = "aws iam list-account-aliases --query AccountAliases[*] --output text"
        elif PROVIDER(self.provider) == PROVIDER.AZURE:
            cmd = "az account show --query name --output tsv"

        output = None

        try:
            output = subprocess.check_output(cmd.split(), stderr=subprocess.STDOUT, universal_newlines=True)

        except subprocess.CalledProcessError as e:
            raise click.ClickException('error fetching {} account alias.\n{}'.format(self.provider, e.output))
            

        except OSError as e:
            raise click.ClickException('error fetching {} account alias.\n{}'.format(self.provider, e.output))

        return output.strip()

    def __write_to_cache(self, k, v):
        """
        Write cache to fs

        Parameters
        ----------
        k:str
            Key
        v:str
            Value
        """
        self.cache[k] = v
        _cache = os.path.join(os.getcwd(), '.tfmake', 'cache')

        with open(_cache, 'w+') as f:
            yaml.safe_dump(self.cache, f, encoding='utf-8', allow_unicode=True)

    def __read_from_cache(self, k):
        """
        Get value from cache

        Parameters
        ----------
        k:str
            Key to retrieve from cache

        Returns
        -------
        str
            Value for 'k' in cache or None
        """
        return self.cache.get(k, None)

    @check_latest_version
    @before_and_after
    def call(self, target, args, dry_run, workspace_key_prefix, **kwargs):
        """
        Call provider specific Makefile using target and (optional) args.

        Parameters
        ----------
        target:str
            Makefile target
        args:list(str)
            Arguments for 'Makefile' target
        dry_run:bool
            When 'True', do _not_ execute 'target'
        workspace_key_prefix:str
            Prefix to use in 'terraform' state

        Raises
        ------
        ClickException
            When 'Makefile' target return exit!=0
        """

        if target == 'env':
            # only execute before+after (because of wrapper)
            return

        makefile = os.path.join(os.path.dirname(os.path.realpath(__file__)), '../Makefile.{}'.format(self.provider))

        if not os.path.isfile(makefile):
            click.secho("Makefile '{}' not found :(".format(makefile), fg='red', bold=True)
            os.sys.exit(1)

        _args = list(args)

        # adding prefix '-' for each argument
        # why? because 'click' does _not_ allow a dash in front of an argument value (~ it thinks its an option)
        # 
        # extra: _only_ when arguments are _not_ meant as terraform arguments (ugly ... i know)
        if target in ['select', 'import']:
            _args=' '.join(_args)
        else:
            if 'TFMAKE_APPROVE' in self.tfmake_env and target in ['apply','destroy'] and not 'auto-approve' in _args:
                _args.append('auto-approve')

            # TODO: handle 'terraform plan' files in a better way ... this is f*cking ugly
            ## @START::ugly
            for arg in _args:
                if arg.endswith('.plan'):
                    # move '.plan' to last in list
                    _args.append(_args.pop(_args.index(arg)))

            # prepare terraform arguments: making sure that 'plan' files are not prefixed with a dash (~ `-`)
            _args = "args='{args}'".format(args=' '.join(['-' + arg if not arg.endswith('.plan') or '=' in arg else arg for arg in _args]))
            ## @END::ugly

        if workspace_key_prefix:
            os.environ['TFMAKE_KEY_PREFIX'] = workspace_key_prefix

        if len(_args) > 0:
            if not dry_run:
                self.__eval(os.system("make -f {file} {target} {args}".format(file=makefile, target=target, args=_args)))
            else:
                click.echo("make -f {file} {target} {args}".format(file=makefile, target=target, args=_args))
        else:
            if not dry_run:
                self.__eval(os.system("make -f {file} {target}".format(file=makefile, target=target)))
            else:
                click.echo("make -f {file} {target}".format(file=makefile, target=target))

        # # write to cache
        # self.__write_to_cache(self.env, self.alias)

    def before(self, target, args, dry_run, workspace_key_prefix):
        '''
        Uses configuration to 'prepare' call to Makefile by setting environment variables and 
        executing arbitrary commands.
        '''
        if not self.config: 
            return

        # when auto-switch enabled, switch to target Azure subscription using 'azctx'
        self.env          = self.__get_environment(target, args)
        self.alias        = self.__get_account_alias()
        self.cached_alias = self.__read_from_cache(self.env)

        if self.config.get('auto_switch', False) and PROVIDER(self.provider) == PROVIDER.AZURE:
            click.secho('AutoSwitch enabled for Azure subscriptions (using cached subscription for {} environment)'.format(self.env), bold=True)
            if self.cached_alias:
                try:
                    from shutil import which
                    # First, check if 'azctx' is installed
                    azctx = which('azctx')
                    if azctx is not None and self.__get_account_alias() != self.cached_alias:
                        p = subprocess.run("{} '{}'".format(azctx, self.cached_alias), 
                            shell              = True,
                            universal_newlines = True,
                            capture_output     = True,
                            text               = True
                        )
                        self.switched = (p.returncode == 0)
                        if self.switched:
                            click.echo(p.stdout)
                        else:
                            raise Exception(p.stderr)

                except Exception as e:
                    raise click.ClickException("switching to Azure subscription '{}': {}".format(self.cached_alias, str(e)))
        elif self.cached_alias and self.alias != self.cached_alias and 'CI_JOB_ID' not in os.environ:
            # besides checking if alias has changed, also check if on Gitlab Runner (if so, don't ask to confirm)
            click.confirm("\n[WARNING] You previously used '{}' for provider {}. Now you're using '{}'. Are you sure?".format(self.cached_alias, self.provider, self.alias), abort=True)

        # first, evaluate environment variables
        _environ = dict()
        for e in self.config.get('environment', []) or []:
            k,v = map(str.strip, e.split('='))
            # Value 'v' could still contain hashtag to comment out rest of line
            v = v.split('#',1)[0]
            if v.startswith('$(') and v.endswith(')'):
                try:
                    pattern = "\$\((.*?)\)"
                    cmd = re.search(pattern, v).group(1) # get substring between $(...)

                    output = subprocess.check_output(cmd.split(), stderr=subprocess.STDOUT, universal_newlines=True)
                    _environ[k] = output

                except subprocess.CalledProcessError as e:
                    raise click.ClickException("exception executing '{}':\n\n{}".format(cmd, e.output))

            else:
                _environ[k]=v

        if target == 'env':
            for k,v in _environ.items():
                print(f'export {k}={v}')

        # merge with os.environ
        for k,v in _environ.items():
            os.environ[k] = v

        # next, run 'before' actions
        for pre in self.config.get('before', []) or []:
            self.__eval(os.system(pre))

    def after(self, target, args, dry_run, workspace_key_prefix):
        '''
        Uses configuration to 'cleanup' after call to Makefile by executing arbitrary commands.
        '''
        if not self.config: 
            return

        if self.switched and PROVIDER(self.provider) == PROVIDER.AZURE:
            # When auto-switch enabled (~ and we actually switched), switch back using 'azctx'
            try:
                from shutil import which
                azctx = which('azctx')
                if azctx is not None:
                    p = subprocess.run('{} -'.format(azctx), 
                        shell              = True,
                        universal_newlines = True,
                        capture_output     = True,
                        text               = True
                    )
                    if p.returncode == 0:
                        click.echo(p.stdout)
                    else:
                        raise Exception(p.stderr)

            except Exception as e:
                raise click.ClickException('switching to previous Azure subscription: {}'.format(str(e)))

        # run 'after' actions
        for post in self.config.get('after', []) or []:
            self.__eval(os.system(post))

        # finally, write to cache
        self.__write_to_cache(self.env, self.alias)


import os
import click
from jinja2 import Environment, PackageLoader
import yaml
import json

from tfmake import __version__

from tfmake.custom import DefaultCommandHandler, DefaultCommandGroup

env = Environment(
    loader=PackageLoader('tfmake', 'templates')
)

@click.group(cls=DefaultCommandGroup)
@click.version_option(version=__version__)
def main():
    pass

@main.command(name='azure')
@click.argument('target', default='help')
@click.argument('args', nargs=-1)
@click.option('--dry-run', is_flag=True, default=False)
@click.option('--workspace-key-prefix')
def azure(target:str, args:list, dry_run:bool, workspace_key_prefix:str):
    '''
    Use Azure provider
    '''
    DefaultCommandHandler('azure').call(target, args, dry_run, workspace_key_prefix)

@main.command(name='guess', default_command=True)
@click.argument('target', default='help')
@click.argument('args', nargs=-1)
@click.option('--dry-run', is_flag=True, default=False)
@click.option('--workspace-key-prefix')
def guess(target:str, args:list, dry_run:bool, workspace_key_prefix:str):
    '''
    Default command that guesses what provider you're using.
    '''
    DefaultCommandHandler().call(target, args, dry_run, workspace_key_prefix)

@main.command(name='aws')
@click.argument('target', default='help')
@click.argument('args', nargs=-1)
@click.option('--dry-run', is_flag=True, default=False)
@click.option('--workspace-key-prefix')
def aws(target:str, args:list, dry_run:bool, workspace_key_prefix:str):
    '''
    Use AWS provider.
    '''
    DefaultCommandHandler('aws').call(target, args, dry_run, workspace_key_prefix)

@main.command()
@click.argument('provider', default='aws')
def init(provider:str):
    '''
    Create configuration for provider (migrate if needed).
    '''
    legacy = None
    cwd = os.path.join(os.getcwd(), '.tfmake')
    if os.path.isfile(cwd):
        if click.confirm('Legacy configuration found! Do you want me to migrate?', default=True):
            with open(cwd,'r') as f:
                legacy = f.read()

        os.rename(cwd, '{}.bkp'.format(cwd))

    if not os.path.isdir(cwd):
        os.mkdir(cwd)

    config = os.path.join(cwd, "config")
    if os.path.isfile(config):
        click.confirm('Existing configuration found! Do you want to continue?', abort=True)

    # Get provider-specific template (this is for future use ... not used yet)
    template = env.get_template('tfmake.{}.j2'.format(provider))

    contents = legacy if legacy else template.render(provider=provider)

    with open(config, 'w+') as f:
        f.write(contents)

    if legacy and os.path.isfile('{}.bkp'.format(cwd)):
        if click.confirm('\n\nSuccessfully migrated configuration. Delete backup?', default=True):
            os.remove('{}.bkp'.format(cwd))

    click.echo("\nConfiguration written to '{}'".format(config))

if __name__ == '__main__':
    main()

